/* Include statements. */
#include <stdio.h>
#include <assert.h>
#include "../linked_lists/queue.c"

/* Define constants. */
#define PRINT_INFO_TESTING_START		"Starting tests.\n"
#define PRINT_INFO_TESTING_END			"Done testing.\n"
#define ERROR_TEXT_RANDOM_QUEUE_FAIL_ALLOCATION	"Long queue test failed because of allocation errors!\n"
#define ERROR_SERIALIZATION_TEST_FAILURE	"Failed the serialization test.\n"
#define SOME_TEST_FAILED			"Recheck the tests, one of them failed!\n"
#define INITIAL_QUEUE_LENGTH			0
#define QUEUE_LENGTH_ONE				1
#define QUEUE_LENGTH_LONG_TEST			4096
#define SERIALIZATION_TEST_QUEUE_LENGTH		5

/* Define functions. */
node_t* generate_random_node();
queue_t* generate_long_random_queue(int length);
int queue_count_length(queue_t *queue);
int test_serialization();
int test_queues();

/* Starting point. */
int main() {
	printf(PRINT_INFO_TESTING_START);
	if (test_linked_lists()) {
		printf(SOME_TEST_FAILED);
	}
	printf(PRINT_INFO_TESTING_END);
}

/* This function generates a random number and puts it in a newly allocated node. */
/* Caller has to free memory afterwards, returns NULL on failure. */
node_t* generate_random_node() {
	/* The node to return. */
	node_t* result;
	int *value;

	/* Attempt to allocate the node. */
	result = (node_t*) malloc(sizeof(node_t));
	if (result == NULL)
		return NULL;
	value = (int*) malloc(sizeof(int));
	if (value == NULL) {
		free(result);
		return NULL;
	}
	*value = random();

	/* Store and return. */
	result->next = NULL;
	result->data = value;
	return result;
}

/* This function generates a long list with random values in the nodes, returns NULL on failures. */
/* Length denotes the length of the list. */
queue_t* generate_long_random_queue(int length) {
	/* Sanity check. */
	if (length <= 0)
		return NULL;
	/* Variables. */
	queue_t *queue;
	node_t *current;
	int i;

	/* Allocate the queue. */
	queue = initialize_queue();
	if (queue == NULL)
		return NULL;

	/* Fill the queue with random nodes. */
	/* First node. */
	current = generate_random_node();
	if (current == NULL) {
		free(queue);
		return NULL;
	}
	/* Attempt to enqueue. */
	if (!enqueue(queue, current)) {
		free(current->data);
		free(current);
		free(queue);
		return NULL;
	}

	/* Loop and enqueue the rest. */
	for (i = 1; i < length; i++) {
		/* Generate and attempt to enqueue. */
		current = generate_random_node();
		if (current == NULL || !enqueue(queue, current)) {
			/* Abort. */
			free(current->data);
			free(current);
			clear_queue_data(&queue, NULL);
			return NULL;
		}
	}

	/* Return the result. */
	return queue;
}

/* This function checks the actual length of a queue. */
int queue_count_length(linked_list_t *list) {
	if (list == NULL || list->head == NULL)
		return 0;
	/* Loop and count. */
	node_t *current = list->head;
	int count = 0;
	while (current != NULL) {
		count++;
		current = current->next;
	}
	/* Return the result. */
	return count;
}

/* This function attempts to allocate a list, fill it, then serialize it. */
int test_serialization() {
	/* Local variables. */
	int i, length;
	/* Used to allocate and store a value. */
	int *value;
	/* Used to store the array. */
	void **array;

	/* Allocations. */
	/* Initialize the queue. */
	queue_t *queue = initialize_queue();
	if (list == NULL) {
		return 1;
	}

	/* Fill the queue. */
	for (i = 0; i < SERIALIZATION_TEST_LIST_LENGTH; i++) {
		/* Allocate the value store. */
		value = (int*) malloc(sizeof(int));
		if (value == NULL) {
			/* Clear the queue. */
			clear_queue_data(&queue, NULL);
			/* Free the array. */
			free(array);
			/* Failed the test, technical reasons. */
			return 1;
		}
		/* Set the value. */
		*value = i;
		/* Enqueue the data. */
		enqueue_data(queue, value);
	}

	/* Serialize the list. */
	array = serialize_queue(queue, &length);
	/* Check if the serialization failed. */
	if (length == -1 || length != list->length) {
		/* Clear the list. */
		clear_queue_data(&queue, NULL);
		/* Free the array. */
		free(array);
		/* Failed the test. */
		return 0;
	}

	/* Check the values. */
	for (i = 0; i < SERIALIZATION_TEST_LIST_LENGTH; i++) {
		if (*((int *)(array[i])) != i) {
			/* Clear the list. */
			clear_queue_data(&queue, NULL);
			/* Free the array. */
			free(array);
			/* Failed the test. */
			return 0;
		}
	}

	/* Passed the tests. */
	/* Clear the list. */
	clear_queue_data(&queue, NULL);
	/* Free the array. */
	free(array);

	return 1;
}

/* This function does the testing. */
int test_queues() {
	/* Local variables. */
	queue_t *queue;
	node_t *node, *tail;
	int error;

	/* Initialize an empty list. */
	queue = initialize_queue();
	/* Check that it was correctly created. */
	assert(queue != NULL);
	assert(queue->length == INITIAL_QUEUE_LENGTH);
	assert(queue->list != NULL);
	assert(queue->list->head == NULL);
	assert(queue->tail == NULL);

	/* Generate and store the node. */
	node = generate_random_node();

	/* Attempt to add it to the queue. */
	enqueue(queue, node);
	/* Check that the queue grew by one and the head got set. */
	assert(queue->list->length == QUEUE_LENGTH_ONE);
	assert(queue->list->head != NULL && queue->list->head == node);
	assert(queue->tail != NULL && queue->tail == node);
	assert(get_node_i_single_linked_list(queue->list, 0, &error) == node);
	assert(get_node_index_pointer_single_linked_list(queue->list, node) == 0);

	/* Attempt to free the nodes and then the list. */
	node = dequeue(queue);
	assert(queue->list->length == INITIAL_QUEUE_LENGTH);
	assert(queue->list->head == NULL);
	assert(queue->tail == NULL);
	node = NULL;
	free(list);
	list = NULL;

	/* Attempt to allocate a long list. */
	list = generate_long_random_list(QUEUE_LENGTH_LONG_TEST);
	if (list == NULL) {
		printf(ERROR_TEXT_RANDOM_QUEUE_FAIL_ALLOCATION);
		return 1;
	}
	assert(queue->list->length == QUEUE_LENGTH_LONG_TEST);
	assert(queue->list->length == queue_count_length(queue));
	/* Test the freeing functions. */
	clear_queue_data(&queue, NULL);
	assert(queue == NULL);

	/* Serialization test. */
	if (!test_serialization()) {
		printf(ERROR_SERIALIZATION_TEST_FAILURE);
		return 1;
	}

	/* Ran successfully, return 0. */
	return 0;
}
