/*
* C_lib Copyright (C) 2021 Wael Karram.
*
* This program is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, version 3 of the License.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

/* Includes. */
#include "max_heap.h"

/* Function implementations. */

/* This function initializes and returns a new EMPTY heap. Returning the pointer thereof on success, or NULL on failure. */
static inline max_heap_t* create_max_heap() {
	/* Attempt to allocate and return accordingly. */
	max_heap_t *result = (max_heap_t*) malloc(sizeof(max_heap_t));
	if (result == NULL)
		return NULL;
	/* Set the fields. */
	result->root = NULL;
	result->size = 0;
	/* Return the result. */
	return result;
}

/* This function attempts to create a new heap and store the data at the root node. Returns a pointer to the heap on success, otherwise NULL. */
static inline max_heap_t* initialize_max_heap_node_store(void *data) {
	/* Local variables. */
	max_heap_t *result;
	heap_node_t *node;

	/* Attempt to allocate the heap. */
	result = create_max_heap();
	if (result == NULL)
		return NULL;

	/* Allocate the node. */
	node = allocate_heap_node(data);
	if (node == NULL) {
		free(result);
		return NULL;
	}

	/* Initialize the heap. */
	result->root = node;
	result->size++;

	/* Return the pointer. */
	return result;
}

/* This function "peeks" at the maximum of the heap. */
/* Returns NULL on errornous input, otherwise returns the max node. */
static inline heap_node_t* find_max_max_heap(max_heap_t *heap) {
	/* Check and return. */
	return (heap != NULL) ? heap->root : NULL;
}

/* Extract the max of the heap. */
static inline void* extract_max_max_heap(max_heap_t *heap) {
	/* Local variables. */
	void *data;

	/* Sanity check. */
	if (!heap || !(heap->root))
		return NULL;

	/* Get the data pointer. */
	data = heap->root->data;

	/* Delete the root. */
	delete_max_max_heap(heap);

	/* Return the data pointer. */
	return data;
}

/* This function replaces the data in the heap's root. */
/* Might exhibit endefined behavior if heap is NULL or has no root. */
static inline void replace_root_max_heap(max_heap_t *heap, void *data) {
	heap->root->data = data;
}

/* This function returns the size of the heap, will return -1 on error! */
static inline size_t size_max_heap(max_heap_t *heap) {
	return (heap == NULL) ? MAX_HEAP_CHECK_ERROR : heap->size;
}

/* This function returns true or false according to if the heap is empty or not, returns MAX_HEAP_CHCEK_ERROR when the state is undefined. */
/* Expect undefined behaviour if size < 0. */
static inline int is_empty_max_heap(max_heap_t *heap) {
	/* Rely on the size function. */
	size_t size = size_max_heap(heap);
	/* Check and return. */
	return (size != MAX_HEAP_CHECK_ERROR) ? size > 0 : MAX_HEAP_CHECK_ERROR;
}

/* This function frees a whole heap, without freeing the data. */
static inline void free_max_heap(max_heap_t *heap) {
	/* Check if there is anything to do. */
	if (heap == NULL)
		return;

	/* Recursively free the inner heap's nodes if needed. */
	if (heap->size && heap->root)
		recursive_free_heap_node(heap->root);
	/* Free the heap struct. */
	free(heap);
}

/* This function frees a whole heap, while also freeing the data. */
/* Will silently fail if free_function is NULL. */
static inline void free_max_heap_data(max_heap_t *heap, void (*free_function)(const void*)) {
	/* Check if there is anything to do. */
	if (!heap || !free_function)
		return;

	/* Recursively free the inner heap's nodes if needed. */
	if (heap->size && heap->root)
		recursive_free_heap_node_data(heap->root, free_function);
	/* Free the heap struct. */
	free(heap);
}

/* This function finds the heap's depth at the deepest leaf. */
/* Will return -1 if the heap is empty (I.E.: root is NULL). */
static inline size_t find_max_heap_depth(max_heap_t *heap) {
	/* Sanity check. */
	if (!heap)
		return (size_t) MAX_HEAP_CHECK_ERROR;

	/* Delegate. */
	return calculate_max_heap_depth(heap->root, (size_t) 0);
}

/* Internal functions. */
/* This function moves a node upwards to its location. */
void sift_up_max_heap(max_heap_t *heap, tree_node_t *node, int (*comparison_function)(const void*, const void*)) {
	/* Keep swapping with the parent until it is in the right place. */
}

/* This function allocates a new heap node, with all pointers to other nodes set to NULL. Returns NULL on failure. */
/* Fills it with the given data - I.E.: points to said data. */
static inline heap_node_t* allocate_heap_node(void *data) {
	/* Variables. */
	heap_node_t *node;

	/* Attempt to allocate the node. */
	node = (heap_node_t*) malloc(sizeof(tree_node_t));
	if (node == NULL)
		return NULL;

	/* Store the data. */
	node->data = data;

	/* Set parent and children to NULL. */
	node->parent = NULL;
	node->right = NULL;
	node->left = NULL;

	/* Return the pointer. */
	return node;
}

/* This function recursively frees heap nodes without freeing the data thereof. */
/* Doesn't do much sanity checking! */
static inline void recursive_free_heap_node(heap_node_t *root) {
	/* Base case. */
	if (!root)
		return;

	/* Recursive case. */
	/* Left. */
	if (root->left)
		recursive_free_heap_node(root->left);
	/* Right. */
	if (root->right)
		recursive_free_heap_node(root->right);

	/* Current node. */
	free(root);
}

/* This function recursively frees heap nodes while freeing the data thereof. */
/* Doesn't do much sanity checking! But will fail if free_function is NULL. */
static inline void recursive_free_heap_node_data(heap_node_t *root, void (*free_function)(const void*)) {
	/* Base case. */
	if (!root)
		return;

	/* Recursive case. */
	/* Left. */
	if (root->left)
		recursive_free_heap_node(root->left);
	/* Right. */
	if (root->right)
		recursive_free_heap_node(root->right);

	/* Current node. */
	free_function(root->data);
	free(root);
}

/* This function recursively looks for the max depth in the heap. */
static inline size_t calculate_max_heap_depth(heap_node_t *root, size_t max) {
	/* Local variables. */
	size_t left, right;

	/* Base case. */
	if (!root)
		return max;

	/* Recursive calls. */
	left = (!root->left) ? (size_t) -1L : calculate_max_heap_depth(root->left);
	right = (!root->right) ? (size_t) -1L : calculate_max_heap_depth(root->right);

	/* Return the max. */
	return MAX(left, MAX(right, max));
}
